/***********************************************************************
 case-control.cc

 This code is part of the source to HapMotif package for analyzing
 conserved regions in haploid sequences of genetic variations.  See the
 README files for more information on the methods, use of the programs,
 and references.  This source code and any associated files are released
 under the terms of the MIT License, reproduced below.

 Copyright (c) 2003 Russell Schwartz, Department of Biological
 Sciences, Carnegie Mellon University.

 Permission is hereby granted, free of charge, to any person obtaining
 a copy of this software and associated documentation files (the
 "Software"), to deal in the Software without restriction, including
 without limitation the rights to use, copy, modify, merge, publish,
 distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to
 the following conditions:

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
************************************************************************/

/***********************************************************************
$Date: 2003/07/15 22:04:53 $
$Revision: 1.4 $
$Log: case-control.cc,v $
Revision 1.4  2003/07/15 22:04:53  russellschwartz

Fixed some ambiguous function calls.

Revision 1.3  2003/07/15 17:42:20  cvsuser
Editing comments.

Revision 1.2  2003/06/30 21:24:24  cvsuser
Adding some more files.

Revision 1.1  2003/06/23 22:17:37  cvsuser
Submitting a first pass at motif-based case-control association testing code.

************************************************************************/

#include <cstdio>
#include <cstdlib>
#include <list>
#include <vector>
#include <map>
#include <algorithm>
#include <math.h>
#include <iostream>
#include <fstream>
#include "haplotype.h"
#include "motif.h"
#include "motiflib.h"
#include "case-control.h"

// A number that is supposed to be much smaller than the
// log of any number greater than zero.
const double LOGZERO=-1e12;

/*************************************************************************
 Description: Find the optimal interpretation of a haplotype in terms of
 motifs.
 Input: haps, the array of all haplotypes; h, the specific haplotype being
 analyzed; mutProb, the mutation probability; motifProbs, a map from
 candidate motif indicies to the log fraction of sequences containing that
 candidate motif; motifs, a map of all motifs listing which are defined;
 maxLength, the maximum allowed motif length; finder, a 2D array of lists,
 indexing motifs by length-1 and starting position
 Output: parse, an ordered list of the motifs making up the optimal parse
 of h
 Returns: the log probability of the optimal parse of h
**************************************************************************/
double parseHaplotype (Haplotype &h, double mutProb,
		       map<Motif,double> &motifProbs, map<Motif,bool> &motifs,
		       list<Motif> &parse, unsigned maxLength,
		       vector<vector<list<Motif> > > &finder)
{
 int i,j;
 int seqlen=h.sequence.size();
 vector<list<Motif> > bestList(seqlen);
 vector<double> bestScores(seqlen);
 vector<int> parents(seqlen);
 vector<Motif> bestMotifs(seqlen);

 // Pick the best starting site
 double opt=LOGZERO;
 Motif optseq;
 int optparent;
 list<Motif>::iterator indices;

 // Iterate over candidate start bases
 for (indices=finder[0][0].begin();
      indices!=finder[0][0].end();
      indices++)
 {
  if (!motifs[*indices])
   continue;

  // Start out with the probability of the given base
  double current=motifProbs[*indices];

  // Add in the score of this base for this sequence to account
  // for possible mutation
  current+=seqScore(h.sequence.substr(0,1),indices->sequence,mutProb);

  // If this is better than previous options, record it
  if (current>opt)
  {
   opt=current;
   optseq=*indices;
  }
 }
 bestScores[0]=opt;
 bestMotifs[0]=optseq;
 parents[0]=-1;

 // Do the dynamic programming iteration
 for (i=1;i<seqlen;i++)
 {
  opt=LOGZERO;
  optparent=-1;
  optseq=Motif();

  // Search over all motifs ending at i
  for (j=i;j>=0 && (unsigned) (i-j)<maxLength;j--)
  {
   for (indices=finder[i-j][j].begin();indices!=finder[i-j][j].end();indices++)
   {
    if (!motifs[*indices])
     continue;

    // Start out with the probability of the given motif
    double current=motifProbs[*indices];

    // Add in the score of this motif for this sequence to account
    // for possible mutation
    current+=seqScore(h.sequence.substr(j,i-j+1),indices->sequence,mutProb);

    // Add in the score of the best previous set of motifs
    if (j>0)
      current+=bestScores[j-1];

    // If this is better than previous options, record it
    if (current>opt)
    {
     opt=current;
     optparent=j-1;
     optseq=*indices;
    }
   }
  }

  // Record the best solution for this value of i
  bestScores[i]=opt;
  bestMotifs[i]=optseq;
  parents[i]=optparent;
 }

 // Backtrack to get the parse
 while (!parse.empty())
  parse.pop_front();
 for (i=seqlen-1;i>=0;i=parents[i])
  parse.push_front(bestMotifs[i]);

 // Return the log of the best probability
 return(bestScores[seqlen-1]);
}

/*************************************************************************
 Description: Print usage information
 Input: argv, the argv argument to main
 Output: prints out a description of the arguments to the program
**************************************************************************/
void usage (char **argv)
{
 cerr << "Usage: " << argv[0] << " -i <case file> <control file> [-s <motif file>] [-m <mutation probability>] [-p <p value>] [-c <min motif frequency>]\n";
 cerr << "-i <case file> <control file>: input sequence file names\n";
 cerr << "-s <motif file>: input motif file name (default: stdin)\n";
 cerr << "-m <mutation probability>: prob of mutation (default: 0)\n";
 cerr << "-p <p value>: p-value for significance test (default: 0.001)\n";
 cerr << "-c <min motif frequency>: frequency cutoff for motifs for association testing (default: 0)\n";
 exit(6);
}

/*************************************************************************
 Description: Inteprets command line arguments.
 Input: argc and argv, the arguments to main
 Output: caseFile and controlFile, the names of files containing
 haplotype data for cases and controls; motifFile, the name of a file
 containing motifs for the case and control data sets, or the empty
 string for standard input; mut, the prior probability of recent mutation;
 pval, the maximum desired p-value for results to be considered
 significant; cutoff, a minimum motif frequency such that motifs above
 below it or above one minus it will not be tested for association
**************************************************************************/
void processCommandLine (int argc, char **argv, string &caseFile,
			 string &controlFile, string &motifFile, double &mut,
			 double &pval, double &cutoff)
{
 int i=1;

 // Set defaults
 caseFile="";
 controlFile="";
 motifFile="";
 mut=0.0;
 pval=0.001;
 cutoff=0.0;
 // Read command line arguments
 while (i<argc)
 {
  string flag(argv[i++]);

  if (flag == "-i")
  {
   if (i>=argc-1)
     usage(argv);
   else
   {
    caseFile=string(argv[i++]);
    controlFile=string(argv[i++]);
   }
  }
  else if (flag == "-s")
  {
   if (i>=argc)
     usage(argv);
   else
    motifFile=string(argv[i++]);
  }
  else if (flag == "-m")
  {
   if (i>=argc)
     usage(argv);
   else
    mut=atof(argv[i++]);
  }
  else if (flag == "-p")
  {
   if (i>=argc)
     usage(argv);
   else
    pval=atof(argv[i++]);
  }
  else if (flag == "-c")
  {
   if (i>=argc)
     usage(argv);
   else
    cutoff=atof(argv[i++]);
  }
  else
   usage(argv);
 }
 if (mut<0.0 || mut>1.0)
 {
  cerr << "Mutation parameter must be between 0 and 1\n";
  usage(argv);
 }
 if (pval<0.0 || pval>1.0)
 {
  cerr << "P-value must be between 0 and 1\n";
  usage(argv);
 }
 if (pval<0.0 || pval>1.0)
 {
  cerr << "Minimum motif frequency must be between 0 and 1\n";
  usage(argv);
 }
 if (caseFile=="" || controlFile=="")
 {
  cerr << "You must specify case and control file names\n";
  usage(argv);
 }
}

/*************************************************************************
 Description: Finds the log of choose(n,k)=n!/(k!(n-k)!) (the number of
 ways of choosing k items from a set of size n).
 Input: n and k, the arguments to choose
 Returns: log(choose(n,k))
**************************************************************************/
double logchoose (int n, int k)
{
 double sum=0.0;
 int i;

 if (k>n || k<0)
  return(1.0);

 for (i=1;i<=k;i++)
  sum-=log((double) i);
 for (i=n-k+1;i<=n;i++)
  sum+=log((double) i);
 return(sum);
}

/*************************************************************************
 Description: Finds the significance of the values in a 2x2 contigency
 table rejecting the hypothesis of independence using Fisher's exact test.
 The test assumes two variables, the first of which is positive for
 cases and negative for controls and the second of which is positive for
 having a given motif and negative for lacking the motif.
 Input: ppo, the count positive for both variables; pno, the count
 positive for the first variable and negative for the second; npo, the
 count negative for the first variable and positive for the second; nno,
 the count negative for both variables
 Returns: the p-value by Fisher's exact test, multiplied by -1 if the
 negative counts for the second variable are over represented among
 those positive for the first variable
**************************************************************************/
double significance  (int ppo, int pno, int npo, int nno)
{
 int n1, n2, k;
 int i;
 bool neg=false;
 double sum=0;

 // Optimization to switch to an equivalent problem if it will
 // reduce the number of results positive in the second variable
 if (ppo+npo>pno+nno)
 {
  ppo^=nno;
  nno^=ppo;
  ppo^=nno;

  npo^=pno;
  pno^=npo;
  npo^=pno;
 }

 // Set up counts of cases, controls, positives
 n1=ppo+pno;
 n2=npo+nno;
 k=ppo+npo;

 // Base p-value on whichever of cases and controls has lower values
 if (((double) ppo)/n1<((double) npo)/n2)
 {
  double current;

  // sum over i up to ppo choose(n1,i)*choose(n2,k-i)/choose(n1+n2,k)
  i=(k>n2) ? (k-n2) : 0;
  current=-exp(logchoose(n1,i)+logchoose(n2,k-i)-logchoose(n1+n2,k));
  sum=current;
  for (;i<ppo;i++)
  {
   current=current*(n1-i)*(k-i)/((i+1)*(n2-k+i+1));
   sum+=current;
  }
 }
 else
 {
  double current;

  // sum over i up to npo choose(n2,i)*choose(n1,k-i)/choose(n1+n2,k)
  i=(k>n1) ? (k-n1) : 0;
  current=exp(logchoose(n2,i)+logchoose(n1,k-i)-logchoose(n1+n2,k));
  sum=current;
  neg=1.0;
  for (;i<npo;i++)
  {
   current=current*(n2-i)*(k-i)/((i+1)*(n1-k+i+1));
   sum+=current;
  }
 }

 // Return the signed significance value
 return(sum);
}

/*************************************************************************
 Description: Finds motifs exceeding a desired significance level in
 distinguishing cases from controls.
 Input: cases and controls, haplotype pairs for cases and controls;
 motifProbs, map from motifs to log probabilities; mut, prior probability
 of recent mutation; pval, maximimum p-value to be considered significant;
 cutoff, minimum motif frequency to be checked
 Output: predictors, map from motifs that are significant predictors of
 case versus control to their p-values, with the p-value set negative for
 predictors of control and positive for predictors of case
**************************************************************************/
void findPredictiveMotifs (vector<pair<Haplotype,Haplotype> > &cases,
 vector<pair<Haplotype,Haplotype> > &controls, map<Motif,double> &motifProbs,
 map<Motif,double> &predictors, double mut, double pval, double cutoff)
{
 vector<vector<list<Motif> > > finder;
 map<Motif,bool> motifs;
 list<Motif> parse;
 unsigned maxLength=0,scale=0;
 map<Motif,int> caseMap, controlMap;
 unsigned i, caseCount=0, controlCount=0;

 // Count hypotheses
 for (map<Motif,double>::iterator mi=motifProbs.begin();mi!=motifProbs.end();
      mi++)
 {
  double p=exp(mi->second);

  if (p>cutoff && 1-p>cutoff)
   scale++;
 }

 // Prepare data structures to facilitate parsing
 finder.resize(cases[0].first.sequence.size());
 for (i=0;i<finder.size();i++)
  finder[i].resize(finder.size());
 for (map<Motif,double>::iterator mi=motifProbs.begin();mi!=motifProbs.end();
      mi++)
 {
  unsigned offset, start;

  // Update finder
  start=mi->first.start;
  offset=mi->first.end-start;
  finder[offset][start].push_front(mi->first);

  // Update maxLength
  if (offset+1>maxLength)
   maxLength=offset+1;

  // Update motifs
  motifs[mi->first]=true;
 }

 // Set up positivity maps for cases
 for (i=0;i<cases.size();i++)
 {
  map<Motif,bool> positives;

  // Find motifs in the current case
  parseHaplotype(cases[i].first,mut,motifProbs,motifs,parse,maxLength,finder);
  for (list<Motif>::iterator pi=parse.begin();pi!=parse.end();pi++)
   positives[*pi]=true;
  parseHaplotype(cases[i].second,mut,motifProbs,motifs,parse,maxLength,finder);
  for (list<Motif>::iterator pi=parse.begin();pi!=parse.end();pi++)
   positives[*pi]=true;
  // Add to counts for any motifs observed
  for (map<Motif,bool>::iterator mpi=positives.begin();mpi!=positives.end();
       mpi++)
   caseMap[mpi->first]+=cases[i].first.count;
  // Add to total count of cases
  caseCount+=cases[i].first.count;
 }

 // Set up positivity maps for controls
 for (i=0;i<controls.size();i++)
 {
  map<Motif,bool> positives;

  // Find motifs in the current control
  parseHaplotype(controls[i].first,mut,motifProbs,motifs,parse,maxLength,
		 finder);
  for (list<Motif>::iterator pi=parse.begin();pi!=parse.end();pi++)
   positives[*pi]=true;
  parseHaplotype(controls[i].second,mut,motifProbs,motifs,parse,maxLength,
		 finder);
  for (list<Motif>::iterator pi=parse.begin();pi!=parse.end();pi++)
   positives[*pi]=true;
  // Add to counts for any motifs observed
  for (map<Motif,bool>::iterator mpi=positives.begin();mpi!=positives.end();
       mpi++)
   controlMap[mpi->first]+=controls[i].first.count;
  // Add to total count of controls
  controlCount+=controls[i].first.count;
 }

 // Test each motif for significance
 cerr << "Testing " << motifProbs.size() << " motifs for significance\n";
 for (map<Motif,double>::iterator mi=motifProbs.begin();mi!=motifProbs.end();
      mi++)
 {
  double sig, tmp, factor1, factor2, siga;
  double p=exp(mi->second);

  // Optionally, skip motifs that are too rare or too frequent
  if (p<=cutoff || 1-p<=cutoff)
   continue;

  // Get significance value
  sig=significance(caseMap[mi->first],caseCount-caseMap[mi->first],
		   controlMap[mi->first],
		   controlCount-controlMap[mi->first]);

  // Correct for multiple hypotheses
  siga=fabs(sig);
  if (2*siga*scale>1.0)
   tmp=(1.0-pow(1.0-siga,(double) scale));
  else
   for (i=0, tmp=0.0, factor1=1.0, factor2=1.0; i<2*scale; i++)
   {
    factor1*=-(2.0*scale-i)*siga/(i+1);
    factor2*=siga/(i+1);
    tmp+=factor2-factor1;
   }
  sig=sig/siga*tmp;

  if (fabs(sig)<pval)
   predictors[mi->first]=sig;
 }
}

int main (int argc, char **argv)
{
 string caseFile,controlFile,motifFile;
 double mut,pval,cutoff;
 map<Motif,double> motifProbs, predictors;
 vector<Haplotype> haplotypes;
 vector<pair<Haplotype,Haplotype> > cases, controls;
 FILE *f;
 unsigned i;

 // Get arguments
 processCommandLine(argc,argv,caseFile,controlFile,motifFile,mut,pval,cutoff);

 // Read motif file
 if (motifFile=="")
 {
  readMotifs(stdin,motifProbs);
 }
 else
 {
  f=fopen(motifFile.c_str(),"r");
  if (f==NULL)
  {
   cerr << "Could not open " << motifFile << " for read.\n";
   exit(1);
  }
  readMotifs(f,motifProbs);
  fclose(f);
 }

 // Read case file
 f=fopen(caseFile.c_str(),"r");
 if (f==NULL)
 {
  cerr << "Could not open " << caseFile << " for read.\n";
  exit(1);
 }
 readHaplotypes(f,haplotypes);
 if (haplotypes.size()%2==1)
 {
  cerr << "Case haplotypes must come in pairs.";
  exit(2);
 }
 cases.resize(haplotypes.size()/2);
 for (i=0;i<cases.size();i++)
 {
  if (haplotypes[2*i].count != haplotypes[2*i+1].count)
  {
   cerr << "Case haplotypes in a pair must have the same multiplicity\n";
   exit(3);
  }
  cases[i]=pair<Haplotype,Haplotype>(haplotypes[2*i],haplotypes[2*i+1]);
 }
 fclose(f);

 // Read control file
 f=fopen(controlFile.c_str(),"r");
 if (f==NULL)
 {
  cerr << "Could not open " << controlFile << " for read.\n";
  exit(1);
 }
 readHaplotypes(f,haplotypes);
 if (haplotypes.size()%2==1)
 {
  cerr << "Control haplotypes must come in pairs.";
  exit(2);
 }
 controls.resize(haplotypes.size()/2);
 for (i=0;i<controls.size();i++)
 {
  if (haplotypes[2*i].count != haplotypes[2*i+1].count)
  {
   cerr << "Control haplotypes in a pair must have the same multiplicity\n";
   exit(3);
  }
  controls[i]=pair<Haplotype,Haplotype>(haplotypes[2*i],haplotypes[2*i+1]);
 }
 fclose(f);

 // Find motifs that distinguish cases from controls
 findPredictiveMotifs(cases,controls,motifProbs,predictors,mut,pval,cutoff);

 // Output the final set of motifs with p-values
 for (map<Motif,double>::iterator pi=predictors.begin();pi!=predictors.end();
      pi++)
 {
  double sig=pi->second;

  if (sig<0.0)
   cout << "Negative predictor ";
  else
   cout << "Positive predictor ";
  cout << "start site: " << pi->first.start << " sequence: " << pi->first.sequence << " significance: " << fabs(pi->second) << "\n";
 }

 return(0);
}
