/***********************************************************************
 haplotype-motif.cc

 This code is part of the source to HapMotif package for analyzing
 conserved regions in haploid sequences of genetic variations.  See the
 README files for more information on the methods, use of the programs,
 and references.  This source code and any associated files are released
 under the terms of the MIT License, reproduced below.

 Copyright (c) 2003 Russell Schwartz, Department of Biological
 Sciences, Carnegie Mellon University.

 Permission is hereby granted, free of charge, to any person obtaining
 a copy of this software and associated documentation files (the
 "Software"), to deal in the Software without restriction, including
 without limitation the rights to use, copy, modify, merge, publish,
 distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to
 the following conditions:

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
************************************************************************/

/***********************************************************************
$Date: 2003/09/15 14:19:08 $
$Revision: 1.10 $
$Log: haplotype-motif.cc,v $
Revision 1.10  2003/09/15 14:19:08  cvsuser
Made correction of known sites an option for prediction code.
Updated READMEs to reflect recent code changes.
Cleaned up and re-commented code in preparation for a new release.

Revision 1.9  2003/08/21 20:24:06  cvsuser
Changed motif-finding model to treat the probability of a motif as
the excess beyond its best explanation in terms of smaller motifs
in the initial discovery phase.

Revision 1.8  2003/08/08 21:08:32  cvsuser
Made various optimizations.  Added a garbage collector to cut down
on memory overhead in htsnp.

Revision 1.7  2003/07/25 14:45:51  cvsuser
Removed unnecessary output and corrected a bug in handling
large haplotype files.

Revision 1.6  2003/06/30 21:24:24  cvsuser
Adding some more files.

Revision 1.5  2003/06/23 17:59:28  cvsuser
Commiting a whole bunch of files because I have not updated them in a
while.

Revision 1.4  2003/05/30 21:06:06  cvsuser
Added code for doing missing site prediction and informative SNP
selection.

Revision 1.3  2003/05/20 19:30:56  cvsuser
Adding some CVS keywords to file.

************************************************************************/

#include <cstdio>
#include <cstdlib>
#include <list>
#include <vector>
#include <map>
#include <algorithm>
#include <math.h>
#include <iostream>
#include <fstream>
#include "haplotype.h"
#include "motif.h"
#include "motiflib.h"
#include "haplotype-motif.h"

// A number that is supposed to be much smaller than the
// log of any number greater than zero.
const double LOGZERO=-1e12;

using namespace std;

/*************************************************************************
 Description: Print usage information
 Input: argv, the argv argument to main
 Output: prints out a description of the arguments to the program
**************************************************************************/
void usage (char **argv)
{
 cerr << "Usage: " << argv[0] << " [-i <input file> ] [-s <motif file>] [-m <mutation probability>] [-o <output format>] [-c confidence] [-x max motif length]\n";
 cerr << "-i <input file>: input file name (default: stdin)\n";
 cerr << "-s <motif file>: optional output file for motif frequencies (default: none)\n";
 cerr << "-m <mutation probability>: prior probability of a base mismatch (default: 0.0)\n";
 cerr << "-o <output format>: either txt (text), pnm (portable anymap), or html (HTML) (default: txt)\n";
 cerr << "-c <confidence>: confidence interval (default: 0.999)\n";
 cerr << "-x <max motif length>: longest motif allowed (default: 20)\n";
 exit(6);
}

/*************************************************************************
 Description: Interpret command line arguments
 Input: argc and argv, the arguments to main
 Output: mut, the mutation probability; inputFile, the file for
 haplotype reference sequences or the empty string if standard input
 is used; outputType, the format of the output, which is one of "txt"
 (text), "pnm" (portable anymap), or "html" (HTML); confidence, the
 size of the confidence interval for use in statistical motif
 validation; maxLength, the length of the longest allowed motif
**************************************************************************/
void processCommandLine (int argc, char **argv, double &mut,
			 string &inputFile, string &outputType,
			 double &confidence, unsigned &maxLength,
			 string &statsFile)
{
 int i=1;

 // Set defaults
 inputFile="";
 outputType="txt";
 statsFile="";
 mut=0.0;
 confidence=0.999;
 maxLength=20;
 // Read command line arguments
 while (i<argc)
 {
  string flag(argv[i++]);

  if (flag == "-i")
  {
   if (i>=argc)
     usage(argv);
   else
    inputFile=string(argv[i++]);
  }
  else if (flag == "-o")
  {
   if (i>=argc)
     usage(argv);
   else
    outputType=string(argv[i++]);
  }
  else if (flag == "-s")
  {
   if (i>=argc)
     usage(argv);
   else
    statsFile=string(argv[i++]);
  }
  else if (flag == "-m")
  {
   if (i>=argc)
     usage(argv);
   else
    mut=atof(argv[i++]);
  }
  else if (flag == "-c")
  {
   if (i>=argc)
     usage(argv);
   else
    confidence=atof(argv[i++]);
  }
  else if (flag == "-x")
  {
   if (i>=argc)
     usage(argv);
   else
    maxLength=atoi(argv[i++]);
  }
  else
   usage(argv);
 }

 if (outputType!="txt" && outputType!="html" && outputType!="pnm")
 {
  cerr << "Invalid output type " << outputType << "\n";
  usage(argv);
 }
 if (mut<0.0 || mut>1.0)
 {
  cerr << "Mutation value " << mut << " must be between 0 and 1.\n";
  usage(argv);
 }
 if (confidence<0.0 || confidence>1.0)
 {
  cerr << "Confidence value " << confidence << " must be between 0 and 1.\n";
  usage(argv);
 }
}

/*************************************************************************
 Description: Perform an initial scan for the frequencies of all potential
 motifs, where any unambiguous substring of less than the maximum length
 found in any sequence is a potential motif.
 Input: haplotypes, the array of all haplotypes; mutProb, the mutation
 probability; devs, the number of standard deviations of a normal
 distribution about the mean needed to establish the chosen confidence
 interval; maxLength, the maximum allowed motif length
 Output: motifProbs, a map from candidate motif indicies to the log
 fraction of sequences containing that candidate motif; motifs, a map of
 all motifs listing which are defined; finder, an 2D array of lists,
 indexing motifs by length-1 and starting position
**************************************************************************/
void findInitialMotifProbs (vector<Haplotype> &haplotypes,
			    map<Motif,double> &motifProbs,
			    map<Motif,bool> &motifs, double mutProb,
			    double devs, unsigned maxLength,
			    vector<vector<list<Motif> > > &finder)
{
 int start,end,seq;
 map<Motif,int> counts;
 map<Motif,int>::const_iterator mi;
 int seqlen=haplotypes[0].sequence.size();
 int total=0;

 // Clean out the motif array
 motifs=map<Motif,bool>();
 finder=vector<vector<list<Motif> > >(maxLength,vector<list<Motif> >(seqlen));

 // Count sequences
 for (seq=0;seq<(int) haplotypes.size();seq++)
  total+=haplotypes[seq].count;

 // Count occurrences of each unique motif
 for (start=0;start<seqlen;start++)
  for (end=start;end<seqlen && (unsigned) (end-start)<maxLength;end++)
  {
   map<string,Motif> seen;

   for (seq=0;seq<(int) haplotypes.size();seq++)
   {
    string s=haplotypes[seq].sequence.substr(start,end-start+1);

    // Reject ambiguous subsequences as possible motifs
    if (ambiguous(s))
     continue;

    // If the motif has already been observed, add to its count.
    // Otherwise, start a new count and record that it has been
    // observed.
    if (seen[s].empty())
    {
     Motif m(start,s);

     finder[end-start][start].push_front(m);
     counts[m]+=haplotypes[seq].count;
     seen[s]=m;
    }
    else
     counts[seen[s]]+=haplotypes[seq].count;
   }
  }

 // Convert occurrence counts to log probabilities
 motifProbs=map<Motif,double>();
 for (mi=counts.begin();mi!=counts.end();mi++)
 {
  motifProbs[mi->first]=log(((double) mi->second)/total);
  motifs[mi->first]=true;
 }
}

/*************************************************************************
 Description: Find the sum of the probabilities of all ways to interpret a
 particular sequence region in terms of motifs.
 Input: haps, the array of all haplotypes; start, the starting site
 of the region under consideration; sequence, the sequence of the region;
 mutProb, the probability of mutation; motifProbs, a map from motif
 indices to log probabilities; motifs, a map telling which motif indices
 correspond to real motifs; maxLength, the longest allowed motif length;
 finder, an 2D array of lists, indexing motifs by length-1 and starting
 position; extra, an extra probability to add to all motif probabilities
 Returns: the log probability of the optimal parse.
**************************************************************************/
double sumRegion (vector<Haplotype> &haps, int start, string sequence,
		  double mutProb, map<Motif,double> &motifProbs,
		  map<Motif,bool> &motifs, unsigned maxLength,
		  vector<vector<list<Motif> > > &finder, double extra)
{
 int i,j;
 int seqlen=haps[0].sequence.size();
 int regionlen=sequence.size();
 vector<list<Motif> > bestList(seqlen);
 vector<double> bestScores(regionlen);
 double opt=0.0;
 list<Motif>::iterator indices;

 // Iterate over candidate start bases
 for (indices=finder[0][start].begin();
      indices!=finder[0][start].end();
      indices++)
 {
  if (!motifs[*indices])
   continue;

  // Start out with the probability of the given base plus the extra
  double current=extra+exp(motifProbs[*indices]);

  // Correct for possible mutations
  current*=exp(seqScore(sequence.substr(0,1),indices->sequence,mutProb));

  // Add to the current total
  opt+=current;
 }
 bestScores[0]=opt;

 // Do the dynamic programming iteration
 for (i=1;i<regionlen;i++)
 {
  opt=0.0;

  // Search over all motifs ending at i
  for (j=i;j>=0 && (unsigned) (i-j)<maxLength && (i-j)<regionlen-1;j--)
  {
   for (indices=finder[i-j][start+j].begin();
	indices!=finder[i-j][start+j].end();
	indices++)
   {
    if (!motifs[*indices])
     continue;

    // Start out with the probability of the given motif
    double current=extra+exp(motifProbs[*indices]);

    // Correct for possible mutations
    current*=exp(seqScore(sequence.substr(j,i-j+1),indices->sequence,mutProb));

    // Multiply by probability of the prefix
    if (j>0)
     current*=bestScores[j-1];

    // Accumulate into current sum
    opt+=current;
   }
  }

  // Store the probability up to base i
  bestScores[i]=opt;
 }

 // Return the log of the best probability
 return(log(bestScores[regionlen-1]));
}

/*************************************************************************
 Description: Find the optimal way to interpret a particular sequence
 region in terms of motifs or single bases it contains.
 Input: haps, the array of all haplotypes; start, the starting site
 of the region under consideration; sequence, the sequence of the region;
 mutProb, the probability of mutation; motifProbs, a map from motif
 indices to log probabilities; motifs, a map telling which motif indices
 correspond to real motifs; maxLength, the longest allowed motif length;
 finder, an 2D array of lists, indexing motifs by length-1 and starting
 position
 Output: parse, the sequence of motifs optimally describing the region
 in question
 Returns: the log probability of the optimal parse.
**************************************************************************/
double parseRegion (vector<Haplotype> &haps, int start, string sequence,
		    double mutProb, map<Motif,double> &motifProbs,
		    map<Motif,bool> &motifs, list<Motif> &parse,
		    unsigned maxLength, vector<vector<list<Motif> > > &finder)
{
 int i,j;
 int seqlen=haps[0].sequence.size();
 int regionlen=sequence.size();
 vector<list<Motif> > bestList(seqlen);
 vector<double> bestScores(regionlen);
 vector<int> parents(seqlen);
 vector<Motif> bestMotifs(seqlen);

 // Pick the best start
 double opt=LOGZERO;
 Motif optseq;
 int optparent;
 list<Motif>::iterator indices;
 // Iterate over candidate start bases
 for (indices=finder[0][start].begin();
      indices!=finder[0][start].end();
      indices++)
 {
  if (!motifs[*indices])
   continue;

  // Start out with the probability of the given base
  double current=motifProbs[*indices];

  // Add in the score of this base for this sequence to account
  // for possible mutation
  current+=seqScore(sequence.substr(0,1),indices->sequence,mutProb);

  // If this is better than previous options, record it
  if (current>opt)
  {
   opt=current;
   optseq=*indices;
  }
 }
 bestScores[0]=opt;
 bestMotifs[0]=optseq;
 parents[0]=-1;

 // Do the dynamic programming iteration
 for (i=1;i<regionlen;i++)
 {
  opt=LOGZERO;
  optparent=-1;
  optseq=Motif();

  // Search over all motifs ending at i
  for (j=i;j>=0 && (unsigned) (i-j)<maxLength && (i-j)<regionlen-1;j--)
  {
   for (indices=finder[i-j][start+j].begin();
	indices!=finder[i-j][start+j].end();
	indices++)
   {
    if (!motifs[*indices])
     continue;

    // Start out with the probability of the given motif
    double current=motifProbs[*indices];

    // Add in the score of this motif for this sequence to account
    // for possible mutation
    current+=seqScore(sequence.substr(j,i-j+1),indices->sequence,mutProb);

    // Add in the score of the best previous set of motifs
    if (j>0)
     current+=bestScores[j-1];

    // If this is better than previous options, record it
    if (current>opt)
    {
     opt=current;
     optparent=j-1;
     optseq=*indices;
    }
   }
  }

  // Record the best solution for this value of i
  bestScores[i]=opt;
  bestMotifs[i]=optseq;
  parents[i]=optparent;

 }

 // Backtrack to get the parse
 while (!parse.empty())
  parse.pop_front();
 for (i=regionlen-1;i>=0;i=parents[i])
  parse.push_front(bestMotifs[i]);

 // Return the log of the best probability
 return(bestScores[regionlen-1]);
}

/*************************************************************************
 Description: Delete any motif whose frequency does not exceed the upper
 confidence bound on its most probable explanation in terms of smaller
 motifs.
 Input: haplotypes, the array of all haplotypes;motifProbs, a map from
 candidate motif indicies to the log fraction of sequences containing that
 candidate motif; motifs, a map of all motifs listing which are defined;
 maxLength, the maximum allowed motif length; mutProb, the mutation
 probability; devs, the number of standard deviations of a normal
 distribution about the mean needed to establish the chosen confidence
 interval; finder, a 2D array of lists,
 indexing motifs by length-1 and starting position
 Output: motifs, updated to reflect deleted motifs; finder, updated to
 reflect deleted motifs
**************************************************************************/
void removeInitialInsignificantMotifs (vector<Haplotype> &haplotypes,
				map<Motif,double> &motifProbs,
				map<Motif,bool> &motifs, unsigned maxLength,
				double mutProb, double devs,
				vector<vector<list<Motif> > > &finder)
{
 unsigned hap,start,offset;
 unsigned seqlen=haplotypes[0].sequence.size();
 unsigned total=0;

 // Count haplotypes
 for (hap=0;hap<haplotypes.size();hap++)
  total+=haplotypes[hap].count;

 // For each motif, see if it is significant
 for (offset=1;offset<seqlen && offset<maxLength;offset++)
 {
  vector<list<Motif> > newFinder(seqlen);

  for (start=0;start+offset<seqlen;start++)
  {
   list<Motif>::iterator indices;

   for (indices=finder[offset][start].begin();
	indices!=finder[offset][start].end();
	indices++)
   {
    Motif index=*indices;
    list<Motif> parse;
    double best,actual;

    if (!motifs[index])
     continue;

    // Find the best explanation of the current motif in terms of
    // smaller motifs
    actual=motifProbs[index];
    best=parseRegion(haplotypes,start,index.sequence,
     mutProb,motifProbs,motifs,parse,maxLength,finder);

    // Compare actual frequency to upper confidence bound on best
    // explanation
    if (actual<upperBound(best,total,devs))
    {
     // Motif is not significant - delete it
     motifs[index]=false;
     motifProbs[index]=LOGZERO;
    }
    else
    {
     // Motif is significant: treat its frequency as the excess of
     // measured frequency over its best explanation in terms of
     // smaller motifs
     motifProbs[index]=log(exp(motifProbs[index])-exp(best));
     // Add to the new lists of motifs with the current offset 
     newFinder[start].push_front(index);
    }
   }
  }
  // Update the finder for this offset value
  finder[offset]=newFinder;
 }
}

/*************************************************************************
 Description: Delete any motif whose frequency does not exceed the upper
 confidence bound on the sum of the probabilities of all possible
 explanations of it in terms of smaller motifs.
 Input: haplotypes, the array of all haplotypes;motifProbs, a map from
 candidate motif indicies to the log fraction of sequences containing that
 candidate motif; motifs, a map of all motifs listing which are defined;
 maxLength, the maximum allowed motif length; mutProb, the mutation
 probability; devs, the number of standard deviations of a normal
 distribution about the mean needed to establish the chosen confidence
 interval; finder, a 2D array of lists,
 indexing motifs by length-1 and starting position
 Output: motifs, updated to reflect deleted motifs; finder, updated to
 reflect deleted motifs
**************************************************************************/
void removeInsignificantMotifs (vector<Haplotype> &haplotypes,
				map<Motif,double> &motifProbs,
				map<Motif,bool> &motifs, unsigned maxLength,
				double mutProb, double devs,
				vector<vector<list<Motif> > > &finder)
{
 unsigned hap,start,offset;
 unsigned seqlen=haplotypes[0].sequence.size();
 unsigned total=0;

 // Count haplotypes
 for (hap=0;hap<haplotypes.size();hap++)
  total+=haplotypes[hap].count;

 // For each motif, see if it is significant
 for (offset=1;offset<seqlen && offset<maxLength;offset++)
 {
  vector<list<Motif> > newFinder(seqlen);

  for (start=0;start+offset<seqlen;start++)
  {
   list<Motif>::iterator indices;

   for (indices=finder[offset][start].begin();
	indices!=finder[offset][start].end();
	indices++)
   {
    Motif index=*indices;
    list<Motif> parse;
    double best,actual;

    if (!motifs[index])
     continue;

    // Find the best explanation of the current motif in terms of
    // smaller motifs
    actual=motifProbs[index];
    best=sumRegion(haplotypes,start,index.sequence,mutProb,motifProbs,motifs,
		   maxLength,finder,exp(actual));

    // Compare actual frequency to upper confidence bound on best
    // explanation
    if (best>LOGZERO && actual<upperBound(best,total,devs))
    {
     // Motif is not significant - delete it
     motifs[index]=false;
     motifProbs[index]=LOGZERO;
    }
    else
    {
     // Add to the new lists of motifs with the current offset 
     newFinder[start].push_front(index);
    }
   }
  }
  // Update the finder for this offset value
  finder[offset]=newFinder;
 }
}

/*************************************************************************
 Description: Count the frequency of each motif given the current
 parses.
 Input: haplotypes, the array of all haplotypes;motifProbs, a map from
 candidate motif indicies to the log fraction of sequences containing that
 candidate motif; motifs, a map of all motifs listing which are defined;
 maxLength, the maximum allowed motif length; mutProb, the mutation
 probability; devs, the number of standard deviations of a normal
 distribution about the mean needed to establish the chosen confidence
 interval; finder, a 2D array of lists, indexing motifs by length-1 and
 starting position
 Output: motifs, updated to reflect deleted motifs; finder, updated to
 reflect deleted motifs
**************************************************************************/
void findMotifProbs (vector<Haplotype> &haplotypes,
		     vector<list<Motif> > &motifParses,
		     map<Motif,double> &motifProbs,
		     map<Motif,bool> &motifs)
{
 unsigned i;
 map<Motif,int> counts;
 map<Motif,int>::const_iterator mi;
 int total=0;

 // Count haplotypes
 for (i=0;i<haplotypes.size();i++)
  total+=haplotypes[i].count;

 // Clean out all observed motifs
 motifs=map<Motif,bool>();

 // count sequences assigned to each motif
 for (i=0;i<motifParses.size();i++)
 {
  list<Motif>::const_iterator pl;

  for (pl=motifParses[i].begin();pl!=motifParses[i].end();pl++)
   counts[*pl]+=haplotypes[i].count;
 }

 // Convert counts to log probabilities
 motifProbs=map<Motif,double>();
 for (mi=counts.begin();mi!=counts.end();mi++)
 {
  double p=((double) mi->second)/total;

  motifProbs[mi->first]=log(p);
  motifs[mi->first]=true;
 }
}

/*************************************************************************
 Description: Find the optimal interpretation of a haplotype in terms of
 motifs.
 Input: haps, the array of all haplotypes; h, the specific haplotype being
 analyzed; mutProb, the mutation probability; motifProbs, a map from
 candidate motif indicies to the log fraction of sequences containing that
 candidate motif; motifs, a map of all motifs listing which are defined;
 maxLength, the maximum allowed motif length; finder, a 2D array of lists,
 indexing motifs by length-1 and starting position
 Output: parse, an ordered list of the motifs making up the optimal parse
 of h
 Returns: the log probability of the optimal parse of h
**************************************************************************/
double parseHaplotype (vector<Haplotype> &haps, Haplotype &h, double mutProb,
		       map<Motif,double> &motifProbs, map<Motif,bool> &motifs,
		       list<Motif> &parse, unsigned maxLength,
		       vector<vector<list<Motif> > > &finder)
{
 int i,j;
 int seqlen=h.sequence.size();
 vector<list<Motif> > bestList(seqlen);
 vector<double> bestScores(seqlen);
 vector<int> parents(seqlen);
 vector<Motif> bestMotifs(seqlen);

 // Pick the best starting site
 double opt=LOGZERO;
 Motif optseq;
 int optparent;
 list<Motif>::iterator indices;

 // Iterate over candidate start bases
 for (indices=finder[0][0].begin();
      indices!=finder[0][0].end();
      indices++)
 {
  if (!motifs[*indices])
   continue;

  // Start out with the probability of the given base
  double current=motifProbs[*indices];

  // Add in the score of this base for this sequence to account
  // for possible mutation
  current+=seqScore(h.sequence.substr(0,1),indices->sequence,mutProb);

  // If this is better than previous options, record it
  if (current>opt)
  {
   opt=current;
   optseq=*indices;
  }
 }
 bestScores[0]=opt;
 bestMotifs[0]=optseq;
 parents[0]=-1;

 // Do the dynamic programming iteration
 for (i=1;i<seqlen;i++)
 {
  opt=LOGZERO;
  optparent=-1;
  optseq=Motif();

  // Search over all motifs ending at i
  for (j=i;j>=0 && (unsigned) (i-j)<maxLength;j--)
  {
   for (indices=finder[i-j][j].begin();indices!=finder[i-j][j].end();indices++)
   {
    if (!motifs[*indices])
     continue;

    // Start out with the probability of the given motif
    double current=motifProbs[*indices];

    // Add in the score of this motif for this sequence to account
    // for possible mutation
    current+=seqScore(h.sequence.substr(j,i-j+1),indices->sequence,mutProb);

    // Add in the score of the best previous set of motifs
    if (j>0)
      current+=bestScores[j-1];

    // If this is better than previous options, record it
    if (current>opt)
    {
     opt=current;
     optparent=j-1;
     optseq=*indices;
    }
   }
  }

  // Record the best solution for this value of i
  bestScores[i]=opt;
  bestMotifs[i]=optseq;
  parents[i]=optparent;
 }

 // Backtrack to get the parse
 while (!parse.empty())
  parse.pop_front();
 for (i=seqlen-1;i>=0;i=parents[i])
  parse.push_front(bestMotifs[i]);

 // Return the log of the best probability
 return(bestScores[seqlen-1]);
}

/*************************************************************************
 Description: Find the optimal interpretation of a set of haplotypes in
 terms of motifs.
 Input: refs, the array of reference haplotypes to which the motifs refer;
 targets, the array of target haplotypes being parsed; mutProb, the
 mutation probability; motifProbs, a map from candidate motif indicies to
 the log fraction of sequences containing that candidate motif; motifs, a
 map of all motifs listing which are defined; maxLength, the maximum
 allowed motif length; finder, a 2D array of lists, indexing motifs by
 length-1 and starting position
 Output: parses, an array of ordered lists of the motifs making up the
 optimal parses of haplotypes in targets
 Returns: the log probability of the product of all optimal parses of
 haplotypes in targets
**************************************************************************/
double parseAllHaplotypes (vector<Haplotype> &refs, vector<Haplotype> &targets,
			   double mutProb, map<Motif,double> &motifProbs,
			   map<Motif,bool> &motifs,
			   vector<list<Motif> > &parses, unsigned maxLength,
			   vector<vector<list<Motif> > > &finder)
{
 unsigned i;
 double score=0.0;

 parses.resize(targets.size());
 for (i=0;i<targets.size();i++)
  score+=parseHaplotype(refs,targets[i],mutProb,motifProbs,motifs,parses[i],
			maxLength,finder);

 return(score);
}

/*************************************************************************
 Description: Find an interpretation of all haplotypes in a set in terms
 of motifs found in the haplotypes while simultaneously inferring the
 motifs and their population frequencies.
 Input: haps, the array of target haplotypes being parsed; mutProb, the
 mutation probability; devs, the number of standard deviations of a normal
 distribution about the mean needed to establish the chosen confidence
 interval; maxLength, the maximum allowed motif length; finder, a 2D array
 of lists, indexing motifs by length-1 and starting position
 Output: motifProbs, a map from candidate motif indicies to the log
 fraction of sequences containing that candidate motif; motifs, a map of
 all motifs listing which are defined; parses, an array of ordered lists
 of the motifs making up the optimal parses of haplotypes in targets
 Returns: the log probability of the product of all optimal parses of
 haplotypes in targets based on the final inferred probabilities and
 motif assignments
**************************************************************************/
double haplotypeMotifEM (vector<Haplotype> &haps,
			   double mutProb, map<Motif,double> &motifProbs,
			 map<Motif,bool> &motifs,
			 vector<list<Motif> > &parses, double devs,
			 unsigned maxLength)
{
 double score=LOGZERO, old;
 vector<vector<list<Motif> > > finder;

 // Find initial frequency estimate
 findInitialMotifProbs(haps,motifProbs,motifs,mutProb,devs,maxLength,finder);
 removeInitialInsignificantMotifs(haps,motifProbs,motifs,maxLength,mutProb,
	 		          devs,finder);

 // Do EM refinement
 do
 {
  old=score;
  score=parseAllHaplotypes(haps,haps,mutProb,motifProbs,motifs,parses,
			   maxLength,finder);
  findMotifProbs(haps,parses,motifProbs,motifs);
  removeInsignificantMotifs(haps,motifProbs,motifs,maxLength,mutProb,
   	 		    devs,finder);

#ifdef DEBUG
   cerr << "Improved score from " << old << " to " << score << "\n";
#endif
 } while (fabs(score-old)>1e-3);

 // Get final probabilities
 findMotifProbs(haps,parses,motifProbs,motifs);

 return(score);
}

/*************************************************************************
 Description: Output a set of colored haplotypes according to a chosen
 format.
 Input: haplotypes, a set of haplotypes; motifParses, an array of the
 ordered sets of motifs describing the parse of each haplotype; the
 coloring of the haplotypes; outputType, a tag signifying the output
 format
 Output: prints to standard output a representation of the haplotypes and
 their coloring
 Notes: This currently allows three formats: text, PNM, and HTML.
**************************************************************************/
void outputMotifs (vector<Haplotype> &haplotypes,
		   vector<list<Motif> > &motifParses,
		   string &outputType)
{
 vector<vector<int> > newParse(haplotypes.size());
 unsigned i;
 int j;
 list<Motif>::const_iterator li;
 int seqlen=haplotypes[0].sequence.size();
 map<Motif,int> counts;
 map<Motif,int>::iterator ci;
 map<Motif,int> colorMap;
 vector<pair<int,Motif> > sorter;

 // Order motifs by frequency
 for (i=0;i<motifParses.size();i++)
  for (li=motifParses[i].begin();li!=motifParses[i].end();li++)
   if (li->start < li->end)
    counts[*li]+=(li->end-li->start+1);
 sorter.resize(counts.size());
 for (i=0,ci=counts.begin();ci!=counts.end();ci++,i++)
  sorter[i]=pair<int,Motif>(-ci->second,ci->first);
 sort(sorter.begin(),sorter.end());
 for (i=0;i<sorter.size();i++)
  colorMap[sorter[i].second]=i+1;

 // Use parses to label individual bases by motif
 for (i=0;i<haplotypes.size();i++)
 {
  int base=0;

  newParse[i].resize(seqlen);

  for (li=motifParses[i].begin();li!=motifParses[i].end();li++)
  {
   int motifLen=li->end-li->start+1;

   for (j=0;j<motifLen;j++)
    newParse[i][j+base]=colorMap[*li];
   base+=motifLen;
  }
 }

 // Dispatch on appropriate output type
 if (outputType=="txt")
  outputHaplotypesTXT(haplotypes,newParse);
 else if (outputType=="html")
  outputHaplotypesHTML(haplotypes,newParse);
 else if (outputType=="pnm")
  outputHaplotypesPNM(haplotypes,newParse);
 else
 {
  cerr << "Unrecognized output type: " << outputType << "\n";
  exit(12);
 }
}

/*************************************************************************
 Description: Output frequencies of the conserved motifs.
 Input: haplotypes, a set of haplotypes; motifProbs, the log probabilities
 of all motifs
 Output: prints to standard error a representation of the motifs of more
 than one base along with their final probabilities
**************************************************************************/
void outputMotifStatistics (vector<Haplotype> haplotypes,
			    map<Motif,double> motifProbs,
			    string fileName)
{
 map<Motif,double>::const_iterator mi;
 vector<pair<double,Motif> > motifList(motifProbs.size());
 unsigned i;
 int j=0;

 // Sort motifs of more than one base by probability
 motifList.resize(motifProbs.size());
 for (mi=motifProbs.begin(),j=0;mi!=motifProbs.end();mi++)
  motifList[j++]=pair<double,Motif>(mi->second,mi->first);
 try
 {
  ofstream statFile(fileName.c_str());

  for (i=0;i<motifList.size();i++)
  {
   Motif index=motifList[i].second;

   statFile << index.start << " " << index.sequence << " ";
   statFile << exp(motifList[i].first) << "\n";
  }
  statFile.close();
 } catch (...)
 {
  cerr << "Could not write frequency file.\n";
  exit(21);
 }
}

int main (int argc, char **argv)
{
 vector<Haplotype> haplotypes;
 vector<list<Motif> > motifParses;
 double mutationProbability;
 map<Motif,double> motifProbs;
 map<Motif,bool> motifs;
 double confidence;
 string inputFile;
 string outputType;
 string statFile;
 unsigned seqlen,maxLength;
 double devs;

 // Read command line
 processCommandLine(argc,argv,mutationProbability,inputFile,outputType,
  confidence,maxLength,statFile);

 // Read data file
 if (inputFile=="")
 {
  readHaplotypes(stdin,haplotypes);
 }
 else
 {
  FILE *f=fopen(inputFile.c_str(),"r");

  if (f==NULL)
  {
   cerr << "Could not open " << inputFile << " for read.\n";
   exit(1);
  }
  readHaplotypes(f,haplotypes);
  fclose(f);
 }

 if (haplotypes.size()==0)
 {
  cerr << "Failed to read any haplotypes\n";
  exit(2);
 }
 seqlen=haplotypes[0].sequence.size();

 devs=sigmaFromInterval(confidence);

 //  Use EM algorithm to find motifs
 haplotypeMotifEM(haplotypes,mutationProbability,motifProbs,motifs,
		  motifParses,devs,maxLength);

 // Output sequence parsed by motifs
 outputMotifs(haplotypes,motifParses,outputType);

 // Output motif statistics
 if (statFile != "")
  outputMotifStatistics(haplotypes,motifProbs,statFile);

 return(0);
}
