/***********************************************************************
 htsnp.cc

 This code is part of the source to HapMotif package for analyzing
 conserved regions in haploid sequences of genetic variations.  See the
 README files for more information on the methods, use of the programs,
 and references.  This source code and any associated files are released
 under the terms of the MIT License, reproduced below.

 Copyright (c) 2003 Russell Schwartz, Department of Biological
 Sciences, Carnegie Mellon University.

 Permission is hereby granted, free of charge, to any person obtaining
 a copy of this software and associated documentation files (the
 "Software"), to deal in the Software without restriction, including
 without limitation the rights to use, copy, modify, merge, publish,
 distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to
 the following conditions:

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
************************************************************************/

/***********************************************************************
$Date: 2003/09/15 14:19:08 $
$Revision: 1.11 $
$Log: htsnp.cc,v $
Revision 1.11  2003/09/15 14:19:08  cvsuser
Made correction of known sites an option for prediction code.
Updated READMEs to reflect recent code changes.
Cleaned up and re-commented code in preparation for a new release.

Revision 1.10  2003/08/08 21:08:32  cvsuser
Made various optimizations.  Added a garbage collector to cut down
on memory overhead in htsnp.

Revision 1.9  2003/07/30 21:39:20  cvsuser
Made algorithmic changes to htsnp.cc to lower memory overhead.

Revision 1.8  2003/07/25 14:45:51  cvsuser
Removed unnecessary output and corrected a bug in handling
large haplotype files.

Revision 1.7  2003/07/16 23:09:47  cvsuser
Corrected some minor errors in usage output of some executables.

Revision 1.6  2003/06/30 21:24:24  cvsuser
Adding some more files.

Revision 1.5  2003/06/23 17:59:28  cvsuser
Commiting a whole bunch of files because I have not updated them in a
while.

Revision 1.4  2003/05/30 21:06:06  cvsuser
Added code for doing missing site prediction and informative SNP
selection.

************************************************************************/

#include <cstdio>
#include <cstdlib>
#include <list>
#include <vector>
#include <map>
#include <algorithm>
#include <math.h>
#include <string.h>
#include <iostream>
#include <fstream>
#include "haplotype.h"
#include "motif.h"
#include "motiflib.h"
#include "htsnp.h"

// Fudge factor for comparing floating point numbers
const double EPSILON=1e-6;

/*************************************************************************
 Description: Find the expected error of a partial solution at a given
 site.
 Input: snps, a string representing a partial solution with "O"
 representing a non-chosen site and "X" a chosen site, starting at the
 first of the SNPs covered by the motif set in the motifs argument; site,
 the site being examined; maxLength, the maximum motif length; motifs,
 a vector of lists of motifs indexed by starting site; motifProbs, a map
 from motifs to their probabilities (which may be adjusted to lower bounds
 of the measured probabilities)
 Returns: the expected error probability at the site, which is equivalent
 to the sum of the probabilities of all motifs for which the most probable
 motif that shares the same values at observed SNPs has a distinct value
 at the site of interest.
**************************************************************************/
double solutionCostPartial (string snps, int site, int offset, int maxLength,
			    vector<list<Motif> > &motifs,
			    map<Motif,double> &motifProbs)
{
 list<Motif> cover;
 double error=1.0;
 vector<Motif> coverv;
 vector<double> probs;
 int first=-1, last=-1;

 // If the site is actually covered by the solution, say the error is 0.0
 if (snps[site-offset]=='X')
  return(0.0);

 // Find all motifs touching the site
 for (int i=((site-maxLength)<0 ? 0 : (site-maxLength));i<=site;i++)
  for (list<Motif>::iterator li=motifs[i].begin();li!=motifs[i].end();li++)
   if (li->start<=site && li->end>=site)
   {
    if (li->start<first || first<0)
     first=li->start;
    if (li->end>last || last<0)
     last=li->end;
    cover.push_front(*li);
   }

 coverv.resize(cover.size());
 probs.resize(cover.size());
 int i=0,j=0;
 for (list<Motif>::iterator lip=cover.begin();lip!=cover.end();lip++)
 {
  coverv[i]=*lip;
  probs[i]=motifProbs[*lip];
  i++;
 }

 // For each motif, find the most probable motif with which it is
 // consistent given the observed sites
 for (i=0;i<(int) coverv.size();i++)
 {
  int argbest=-1;
  double bestProb=0.0;

  for (j=0;j<(int) coverv.size();j++)
  {
   bool distinguishable=false;

   if (probs[j]<=bestProb)
    continue;

   for (int k=first; k<=last; k++)
    if (snps[k-offset]=='X')
     if (k>=coverv[i].start && k<=coverv[i].end && k>=coverv[j].start &&
	 k<=coverv[j].end)
      if (coverv[i].sequence[k-coverv[i].start]!=
	  coverv[j].sequence[k-coverv[j].start])
      {
       distinguishable=true;
       break;
      }

   if (!distinguishable)
   {
    bestProb=probs[j];
    argbest=j;
   }
  }

  // If the best representative is identical at the current site,
  // then record the motif as a non-error
  if (coverv[i].sequence[site-coverv[i].start]==
      coverv[argbest].sequence[site-coverv[argbest].start])
   error-=probs[i];
 }

 return(error);
}

/*************************************************************************
 Description: Increment a binary string of 'O's and 'X's
 Input: last, the previous string
 Output: last, the string incremented by one, where 'O'=0 and 'X'=1.
 If the string is all 'X's, it reverts to all 'O's
 Returns: true if the input contains any 'O's, false otherwise
**************************************************************************/
bool incrementString (string &last)
{
 for (int i=last.size()-1;i>=0;i--)
 {
  if (last[i]=='O')
  {
   last[i]='X';
   return(true);
  }
  else
   last[i]='O';
 }
 return(false);
}

/*************************************************************************
 Description: Given a string, extend it by all possible binary strings
 of 'X's and 'O's of a given length.
 Input: oldS, the original prefix string; extra, the length of the
 extra bits to add
 Returns: a list of all of the extended strings
**************************************************************************/
list<string> extendString (string oldS, int extra)
{
 string newS=string(extra,'O');
 list<string> output;

 do
 {
  output.push_front(oldS + newS);
 } while (incrementString(newS));

 return(output);
}

/*************************************************************************
 Description: Counts the number of 'X's in a string.
 Input: s, the input string
 Returns: the number of 'X's in the string
**************************************************************************/
int countX (string s)
{
 int sum=0;

 for (unsigned i=0;i<s.size();i++)
  if (s[i]=='X')
   sum++;

 return(sum);
}

/*************************************************************************
 Description: Takes a string consisting of an integer, space, and
 whitespace-free string and extracts the integer and string.
 Input: key, the input string
 Returns: a pair containing the whitespace-free string and the integer
 Note: this is an auxiliary routine to allow for indexing by both an
 integer and a string using a map
**************************************************************************/
inline pair<string,int> splitHapKey (string key)
{
 char buffer[key.size()+1];
 pair<string,int> ret;
 int r;

 r=sscanf(key.c_str(),"%d %s",&ret.second,buffer);
 if (r==2)
  ret.first=string(buffer);
 else
  ret.first=string("");

 return(ret);
}

/*************************************************************************
 Description: Joins a whitespace-free string and an integer into a
 single string consisting of the integer, a space, and the whitespace-free
 string
 Input: s, the input string; i, the integer
 Returns: the joined string
 Note: this is an auxiliary routine to allow for indexing by both an
 integer and a string using a map
**************************************************************************/
inline string joinHapKey (string s, int i)
{
 char buffer[s.size()+128];
 string ret;

 sprintf(buffer,"%d %s",i,s.c_str());
 ret=string(buffer);

 return(ret);
}

/*************************************************************************
 Description: Given a fixed maximum number of SNPs, choses the subset
 that minimizes expected prediction error.
 Input: motifProbs, a map from motifs to their probabilities; maxSNPs,
 the maximum number of SNPs
 Output: isnps, a list of the chosen SNPs
 Returns: the expected total number of prediction errors
**************************************************************************/
double findBestInformativeSNPs (map<Motif,double> &motifProbs,
				list<int> &isnps, int maxSNPs)
{
 int seqlen=0;
 map<Motif,double>::iterator mi;
 vector<string> previous;
 vector<int> prevCosts;
 vector<int> starts,ends;
 vector<list<Motif> > motifs;
 int oldEnd=-1,maxLength=-1;
 double opt;
 string argopt;

 vector<map<string,double> > solutions;
 vector<map<string,string> > parents;
 vector<map<string,int> > refCounts;

 // Find the sequence length and max motif length
 for (mi=motifProbs.begin();mi!=motifProbs.end();mi++)
 {
  if (mi->first.end>seqlen)
   seqlen=mi->first.end;
  if (mi->first.end-mi->first.start+1>maxLength)
   maxLength=mi->first.end-mi->first.start+1;
 }
 seqlen++;

 starts=vector<int>(seqlen,seqlen);
 ends=vector<int>(seqlen,0);

 // Clean out isnps output array
 while (!isnps.empty())
  isnps.pop_front();

 // Initialize the search
 solutions.resize(seqlen+1);
 solutions[0][joinHapKey(string(""),0)]=0.0;
 parents.resize(seqlen+1);
 parents[0][string("")]=string("");
 refCounts.resize(seqlen+1);
 refCounts[0][string("")]=1;

 // Find the first and last value index of any motif touching each site
 for (map<Motif,double>::iterator mmi=motifProbs.begin();
      mmi!=motifProbs.end();mmi++)
 {
  Motif m=mmi->first;

  for (int i=m.start;i<=m.end;i++)
  {
   if (m.start<starts[i])
    starts[i]=m.start;
   if (m.end>ends[i])
    ends[i]=m.end;
  }
 }

 // Create a data structure of motifs indexed by start site
 motifs.resize(seqlen);
 for (map<Motif,double>::iterator mmi=motifProbs.begin();
      mmi!=motifProbs.end();mmi++)
  motifs[mmi->first.start].push_back(mmi->first);

 // For each site, establish possible solutions
 for (int i=0;i<seqlen;i++)
 {
  int start=starts[i],end=ends[i];
  list<string> allExtensions;
  vector<pair<string,int> > extensions;
  unsigned k;
  map<string,bool> touched;

  // Find extensions and put into vector
  if (end>oldEnd)
   allExtensions=extendString(string(""),end-oldEnd);
  else
   allExtensions.push_front("");
  extensions.resize(allExtensions.size());
  k=0;
  for (list<string>::iterator ei=allExtensions.begin();
       ei!=allExtensions.end();ei++,k++)

   extensions[k]=pair<string,int>(*ei,countX(*ei));

  cerr << "SNP " << i << "\n";

  // Test all possible extensions on all possible prefixes
  for (map<string,double>::iterator si=solutions[i].begin();
       si!=solutions[i].end();si++)
  {
   string oldPrefix=splitHapKey(si->first).first;
   int oldSize=splitHapKey(si->first).second;

   for (k=0;k<extensions.size();k++)
   {
    int newSize=oldSize+extensions[k].second;

    if (newSize<=maxSNPs)
    {
     int offset=((i>0) ? starts[i-1] : 0);
     string current=oldPrefix+extensions[k].first;
     double cost=solutionCostPartial(current,i,offset,
				     maxLength,motifs,motifProbs)+
                 si->second;
     string post=current.substr(start-offset,end-start+1);
     string key=joinHapKey(post,newSize);

     if (!touched[key] || cost<solutions[i+1][key])
     {
      solutions[i+1][key]=cost;
      parents[i+1][key]=si->first;
      touched[key]=true;
     }
    }
   }
  }

  // Set up reference counters
  for (map<string,string>::iterator pi=parents[i+1].begin();
       pi!=parents[i+1].end();
       pi++)
   refCounts[i][pi->second]++;

  // Prune unreferenced data structures
  for (map<string,double>::iterator ri=solutions[i].begin();
       ri!=solutions[i].end();ri++)
  {
   if (refCounts[i][ri->first]==0)
   {
    int level=i;
    string par=ri->first;

    while(refCounts[level][par]==0)
    {
     string parnew=parents[level][par];

     refCounts[level].erase(par);
     solutions[level].erase(par);
     parents[level].erase(par);
     level--;
     refCounts[level][parnew]--;
     par=parnew;
    }
   }
  }

 // Prepare for next round of iteration
  oldEnd=end;
 }

 // Do backtracking

 // Find the best solution from the final round
 opt=seqlen+1.0;
 argopt=string("");
 for (map<string,double>::iterator mi=solutions[seqlen].begin();
      mi!=solutions[seqlen].end();mi++)
  if (mi->second<opt && splitHapKey(mi->first).second==maxSNPs)
  {
   argopt=mi->first;
   opt=mi->second;
  }

 // Backtrack
 list<string> blocks;
 blocks.push_front(splitHapKey(argopt).first);
 int snp=seqlen-2;
 while (snp>=0)
 {
  string nextString;

  argopt=parents[snp+2][argopt];
  if (starts[snp+1]>starts[snp])
   nextString=splitHapKey(argopt).first.substr(0,starts[snp+1]-starts[snp]);
  else
   nextString=string("");
  blocks.push_front(nextString);
  snp--;
 }

 // Put it all together
 argopt="";
 for (list<string>::iterator li=blocks.begin();li!=blocks.end();li++)
  argopt+=*li;

 // Put the SNP set into the isnps list
 for (unsigned j=0;j<argopt.size();j++)
  if (argopt[j]=='X')
   isnps.push_back(j);

 return(opt);
}

/*************************************************************************
 Description: Choose a minimal set of SNPs from which each missing SNP
 can be predicted with a specified accuracy.
 Input: motifProbs, a map from motifs to their probabilities; maxError,
 the maximum allowed per-site error
 Output: isnps, a list of the chosen SNPs
 Returns: the number of SNPs needed
**************************************************************************/
int findMaxInformativeSNPs (map<Motif,double> &motifProbs,
				list<int> &isnps, double maxError)
{
 int seqlen=0;
 map<Motif,double>::iterator mi;
 vector<string> previous;
 vector<int> prevCosts;
 vector<int> starts,ends;
 vector<list<Motif> > motifs;
 int oldEnd=-1,maxLength=-1;
 double opt;
 int iopt;
 string argopt;

 vector<map<string,int> > solutions;
 vector<map<string,double> > costs;
 vector<map<string,string> > parents;
 vector<map<string,int> > refCounts;

 // Find the sequence length and max motif length
 for (mi=motifProbs.begin();mi!=motifProbs.end();mi++)
 {
  if (mi->first.end>seqlen)
   seqlen=mi->first.end;
  if (mi->first.end-mi->first.start+1>maxLength)
   maxLength=mi->first.end-mi->first.start+1;
 }
 seqlen++;

 starts=vector<int>(seqlen,seqlen);
 ends=vector<int>(seqlen,0);

 // Clean out isnps output array
 while (!isnps.empty())
  isnps.pop_front();

 solutions.resize(seqlen+1);
 solutions[0][string("")]=0;
 costs.resize(seqlen+1);
 costs[0][string("")]=0.0;
 parents.resize(seqlen+1);
 parents[0][string("")]=string("");
 refCounts.resize(seqlen+1);
 refCounts[0][string("")]=1;

 vector<map<string,int> > eCount(seqlen+1);

 // Find the first and last value index of any motif touching each site
 for (map<Motif,double>::iterator mmi=motifProbs.begin();
      mmi!=motifProbs.end();mmi++)
 {
  Motif m=mmi->first;

  for (int i=m.start;i<=m.end;i++)
  {
   if (m.start<starts[i])
    starts[i]=m.start;
   if (m.end>ends[i])
    ends[i]=m.end;
  }
 }

 // Create a data structure of motifs indexed by start site
 motifs.resize(seqlen);
 for (map<Motif,double>::iterator mmi=motifProbs.begin();
      mmi!=motifProbs.end();mmi++)
  motifs[mmi->first.start].push_back(mmi->first);

 // For each site, establish possible solutions
 for (int i=0;i<seqlen;i++)
 {
  int start=starts[i],end=ends[i];
  list<string> allExtensions;
  vector<pair<string,int> > extensions;
  unsigned k;
  map<string,bool> touched;

  cerr << "SNP " << i << "\n";

  // Find extensions and put into vector
  if (end>oldEnd)
   allExtensions=extendString(string(""),end-oldEnd);
  else
   allExtensions.push_front("");
  extensions.resize(allExtensions.size());
  k=0;
  for (list<string>::iterator ei=allExtensions.begin();
       ei!=allExtensions.end();ei++,k++)
   extensions[k]=pair<string,int>(*ei,countX(*ei));

  map<string,string> postOpt;
  map<string,int> postSize;
  map<string,double> postCost;
  map<string,bool> seen;
  int matchLen=oldEnd-start+1;
  for (map<string,int>::iterator si=solutions[i].begin();
       si!=solutions[i].end();si++)
  {
   int offset=((i>0) ? starts[i-1] : 0);
   string post=(matchLen>0) ? si->first.substr(start-offset,matchLen) :
               string("");

   if (!seen[post] || postSize[post]>si->second ||
       (postSize[post]==si->second && postCost[post]>costs[i][si->first]))
   {
    postSize[post]=si->second;
    postCost[post]=costs[i][si->first];
    postOpt[post]=si->first;
    seen[post]=true;
   }
  }

  // Test all possible extensions on all possible prefixes
  for (map<string,string>::iterator si=postOpt.begin();
      si!=postOpt.end();si++)
  {
   string oldPrefix=si->second;
   int oldSize=postSize[si->first];

   for (k=0;k<extensions.size();k++)
   {
    int newSize=oldSize+extensions[k].second;
    int offset=((i>0) ? starts[i-1] : 0);
    string current=oldPrefix+extensions[k].first;
    string post=current.substr(start-offset,end-start+1);
    double currentCost=solutionCostPartial(current,i,offset,maxLength,motifs,
     motifProbs);
    double newCost=currentCost+costs[i][oldPrefix];

    if (currentCost<maxError+EPSILON)
     if (!touched[post] || solutions[i+1][post]>newSize ||
         (solutions[i+1][post]==newSize && costs[i+1][post]>newCost))
     {
      solutions[i+1][post]=newSize;
      costs[i+1][post]=newCost;
      parents[i+1][post]=oldPrefix;
      touched[post]=true;
     }
   }
  }

  // Set up reference counters
  for (map<string,string>::iterator pi=parents[i+1].begin();
       pi!=parents[i+1].end();
       pi++)
   refCounts[i][pi->second]++;

  // Prune unreferenced data structures
  for (map<string,int>::iterator ri=solutions[i].begin();
       ri!=solutions[i].end();ri++)
  {
   if (refCounts[i][ri->first]==0)
   {
    int level=i;
    string par=ri->first;

    while(refCounts[level][par]==0)
    {
     string parnew=parents[level][par];

     refCounts[level].erase(par);
     solutions[level].erase(par);
     costs[level].erase(par);
     parents[level].erase(par);

     eCount[level][par]++;
     level--;
     refCounts[level][parnew]--;
     par=parnew;
    }
   }
  }

  // Prepare for next round of iteration
  oldEnd=end;
 }

 // Find the best solution from the final round
 opt=seqlen+1.0;
 iopt=seqlen+1;
 argopt=string("");
 for (map<string,int>::iterator mi=solutions[seqlen].begin();
      mi!=solutions[seqlen].end();mi++)
 {
  if (mi->second<iopt)
  {
   argopt=mi->first;
   iopt=mi->second;
   opt=costs[seqlen][mi->first];
  }
  else if (mi->second==iopt)
   if (costs[seqlen][mi->first]<opt)
   {
    argopt=mi->first;
    iopt=mi->second;
    opt=costs[seqlen][mi->first];
   }
 }

 // Backtrack
 list<string> blocks;
 blocks.push_front(argopt);
 int snp=seqlen-2;
 while (snp>=0)
 {
  string nextString;

  argopt=parents[snp+2][argopt];
  if (starts[snp+1]>starts[snp])
   nextString=argopt.substr(0,starts[snp+1]-starts[snp]);
  else
   nextString=string("");
  blocks.push_front(nextString);
  snp--;
 }

 // Put it all together
 argopt="";
 for (list<string>::iterator li=blocks.begin();li!=blocks.end();li++)
  argopt+=*li;

 // Put the SNP set into the isnps list
 for (unsigned j=0;j<argopt.size();j++)
  if (argopt[j]=='X')
   isnps.push_back(j);

 return(iopt);
}

/*************************************************************************
 Description: Choose a minimal set of SNPs from which missing SNPs can
 on average be predicted with a specified accuracy.
 Input: motifProbs, a map from motifs to their probabilities; avgError,
 the maximum expected prediction error
 Output: isnps, a list of the chosen SNPs
 Returns: the expected average error
**************************************************************************/
double findAvgInformativeSNPs (map<Motif,double> &motifProbs, list<int> &isnps,
			       double avgError)
{
 int min, max, mid, seqlen=-1;
 double total, minscore, maxscore, midscore=0.0;

 for (map<Motif,double>::iterator mi=motifProbs.begin();
      mi!=motifProbs.end();mi++)
  if (mi->first.end>seqlen)
   seqlen=mi->first.end;
 seqlen++;
 total=avgError*seqlen;

 // Find minimum for binary search
 min=0;
 minscore=findBestInformativeSNPs(motifProbs,isnps,0);

 // Find maximum for search
 maxscore=seqlen+1;
 for (max=1;maxscore>total && max<seqlen;max*=2)
  maxscore=findBestInformativeSNPs(motifProbs,isnps,max);
 if (max>seqlen)
  max=seqlen;

 // Do search until solution is found
 while (min<max-1)
 {
  mid=(min+max)/2;
  midscore=findBestInformativeSNPs(motifProbs,isnps,mid);
  if (midscore>total)
  {
   min=mid;
   minscore=midscore;
  }
  else
  {
   max=mid;
   maxscore=midscore;
  }
 }
 if (midscore>total)
  midscore=findBestInformativeSNPs(motifProbs,isnps,max);

 return(midscore/seqlen);
}

/*************************************************************************
 Description: Print usage information
 Input: argv, the argv argument to main
 Output: prints out a description of the arguments to the program
**************************************************************************/
void usage (char **argv)
{
  cerr << "Usages: \n";
 cerr << argv[0] << " [-i <input file> ] [-d <maximum error>] [-p <population> [-c <confidence interval>]]\n";
 cerr << argv[0] << " [-i <input file> ] [-a <average error>] [-p <population> [-c <confidence interval>]]\n";
 cerr << argv[0] << " [-i <input file> ] [-k <max SNPs>]\n";
 cerr << "-i <input file> : specifies a file of motif frequencies to use in SNP selection (default: stdin)\n";
 cerr << "-d <maximum error> : specifies a tolerated maximum amount of per-base error and seeks to minimize SNPs for that maximum (default: 0)\n";
 cerr << "-a <average error> : specifies a tolerated average amount of per-base error and seeks to minimize SNPs for that average (default: 0)\n";
 cerr << "-k <max SNPs> : specifies a maximum number of SNPs and seeks to minimize expected total error given that maximum (default: 0)\n";
 cerr << "-p <population> : specifies the population size from which motifs were derived (only used if a confidence interval is used)\n";
 cerr << "-c <confidence interval> : specifies the size of the confidence interval if a population size is specified (default: 0.0)\n";

 exit(6);
}

/*************************************************************************
 Description: Interpret command line arguments
 Input: argc and argv, the arguments to main
 Output: inputFile, the file for
 motif sequences and probabilities or the empty string if standard input
 is used; maxDiv, the maximum per-site error; population, the size of the
 population used to find the motifs or -1 if no population size is
 specified confidence, the size of the confidence interval for use in
 statistical motif validation
**************************************************************************/
void processCommandLine (int argc, char **argv, string &inputFile,
			 double &maxError, int &population,
			 double &confidence, int &maxSNPs, double &avgError)
{
 int i=1;
 bool ms=false;

 // Set defaults
 inputFile="";
 maxError=0.0;
 population=-1;
 confidence=0.0;
 avgError=0.0;
 maxSNPs=0;
 // Read command line arguments
 while (i<argc)
 {
  string flag(argv[i++]);

  if (flag == "-i")
  {
   if (i>=argc)
     usage(argv);
   else
    inputFile=string(argv[i++]);
  }
  else if (flag == "-d")
  {
   if (i>=argc)
     usage(argv);
   else
    maxError=atof(argv[i++]);
  }
  else if (flag == "-a")
  {
   if (i>=argc)
     usage(argv);
   else
    avgError=atof(argv[i++]);
  }
  else if (flag == "-p")
  {
   if (i>=argc)
     usage(argv);
   else
    population=atoi(argv[i++]);
  }
  else if (flag == "-k")
  {
   if (i>=argc)
     usage(argv);
   else
    maxSNPs=atoi(argv[i++]);
   ms=true;
  }
  else if (flag == "-c")
  {
   if (i>=argc)
     usage(argv);
   else
    confidence=atof(argv[i++]);
  }
  else
   usage(argv);
 }

 if (confidence<0.0 || confidence>1.0)
 {
  cerr << "Confidence interval must be between 0 and 1\n";
  exit(11);
 }
 if (maxError<0.0 || maxError>1.0)
 {
  cerr << "Maximum error value " << maxError << " must be between 0 and 1.\n";
  usage(argv);
 }
 if (avgError<0.0 || avgError>1.0)
 {
  cerr << "Average error value " << avgError << " must be between 0 and 1.\n";
  usage(argv);
 }
 if (maxSNPs<0)
 {
  cerr << "Maximum number of SNPs must be non-negative.\n";
  usage(argv);
 }
 if (ms && maxSNPs==0)
  maxSNPs=-1;
 if (maxSNPs>0 && maxError>0.0)
 {
  cerr << "You can only specify one of maximum SNPs and maximum error.\n";
  usage(argv);
 }
 if (maxSNPs>0 && avgError>0.0)
 {
  cerr << "You can only specify one of maximum SNPs and average error.\n";
  usage(argv);
 }
 if (maxError>0.0 && avgError>0.0)
 {
  cerr << "You can only specify one of maximum error and average error.\n";
  usage(argv);
 }
 if (maxSNPs>0 && population>0)
 {
  cerr << "Population is not meaningful when used with maxSNPs option.\n";
  usage(argv);
 }
 if (confidence>0 && population<0)
 {
  cerr << "You must specify a population size to use confidence bounds.\n";
  usage(argv);
 }
}

int main (int argc, char **argv)
{
 string inputFile;
 double maxError;
 double devs=0.0;
 map<Motif,double> motifProbs;
 map<Motif,double>::iterator mi;
 list<int> isnps;
 list<int>::iterator li;
 int population, maxSNPs;
 double confidence, avgError;

 // Read arguments
 processCommandLine(argc,argv,inputFile,maxError,population,confidence,maxSNPs,
		    avgError);

 // Read motif data file
 if (inputFile=="")
 {
  readMotifs(stdin,motifProbs);
 }
 else
 {
  FILE *f=fopen(inputFile.c_str(),"r");

  if (f==NULL)
  {
   cerr << "Could not open " << inputFile << " for read.\n";
   exit(1);
  }
  readMotifs(f,motifProbs);
  fclose(f);
 }

 // Get number of standard deviations for sample size correction
 devs=sigmaFromInterval(confidence);

 // Convert log probabilities to regular probabilities
 for (mi=motifProbs.begin();mi!=motifProbs.end();mi++)
 {
  if (population<=0)
   motifProbs[mi->first]=exp(mi->second);
  else
   motifProbs[mi->first]=exp(lowerBound(mi->second,population,devs));
 }

 // Find the smallest valid set of SNPs
 if (maxSNPs<0)
 // For some reason, someone wants a size 0 solution
 {
  isnps=list<int>();
 }
 else if (maxSNPs>0)
 // Option 1: For a fixed number of SNPs, find the set that minimizes
 // total expected error 
 {
  double cost=findBestInformativeSNPs(motifProbs,isnps,maxSNPs);

  cerr << "Expected cost is " << cost << "\n";
 }
 else if (avgError>0.0)
 // Option 2: For a desired average error, find the minimal SNP set
 // that does not exceed it (either in expectation or with a given
 // confidence). 
 {
  double cost=findAvgInformativeSNPs(motifProbs,isnps,avgError);

  cerr << "Final average is " << cost << "\n";
 }
 else
 // Option 3: For a desired maximum error, find the minimal SNP set
 // that does not exceed it at any site (either in expectation or
 // with a given confidence). 
  findMaxInformativeSNPs(motifProbs,isnps,maxError);

 // Output the answer
 cerr << "Resolved with " << isnps.size() << " SNPs:";
 for (li=isnps.begin();li!=isnps.end();li++)
  cout << " " << *li;
 cout << "\n";

 return(0);
}
