/***********************************************************************
 motiflib.cc

 This code is part of the source to HapMotif package for analyzing
 conserved regions in haploid sequences of genetic variations.  See the
 README files for more information on the methods, use of the programs,
 and references.  This source code and any associated files are released
 under the terms of the MIT License, reproduced below.

 Copyright (c) 2003 Russell Schwartz, Department of Biological
 Sciences, Carnegie Mellon University.

 Permission is hereby granted, free of charge, to any person obtaining
 a copy of this software and associated documentation files (the
 "Software"), to deal in the Software without restriction, including
 without limitation the rights to use, copy, modify, merge, publish,
 distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to
 the following conditions:

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
************************************************************************/

#include <cstdio>
#include <cstdlib>
#include <list>
#include <vector>
#include <map>
#include <algorithm>
#include <math.h>
#include <iostream>
#include <fstream>
#include "haplotype.h"
#include "motif.h"
#include "motiflib.h"

/*************************************************************************
 Description: Assign a score to a pair of strings reflecting the
 probability the second is a recent ancestor of the first.
 Input: s1 and s2, the two strings; m, the probability of an individual
 base being mutated in s1. 
 Returns: The log probability of s1 being descended from s2 under the model
 that mismatches have probability m, matches have probability 1-m, and
 ambiguous sites contribute no information.
**************************************************************************/
double seqScore (string s1, string s2, double m)
{
 unsigned i;
 double score=0.0;
 double match=log(1-m),mismatch=log(m);

 for (i=0;i<s1.size();i++)
  if (s1[i]=='X' || s1[i]=='x' || s1[i]=='?')
   continue;
  else if (s2[i]=='X' || s2[i]=='x' || s2[i]=='?')
   continue;
  else if (s1[i]==s2[i])
   score+=match;
  else
   score+=mismatch;

 return(score);
}

/*************************************************************************
 Description: Read one line of input from a file.
 Requires: the input file is open for reading
 Input: f, a pointer to the haplotype input file
 Returns: a pointer to a string containing the next line of input
 Notes: It might be better to implement this with a C++ iostream instead.
**************************************************************************/
char *getOneLine (FILE *f)
{
 unsigned size=1024,totalSize=0;
 char *buffer=(char *) malloc(size*sizeof(char)),
      *total=(char *) malloc(size*sizeof(char));
 bool done=false;

 total[0]=0;
 while (!done)
 {
  unsigned len;
  char *tmp;

  tmp=fgets(buffer,size-1,f);
  if (tmp==NULL)
  {
   free(buffer);
   free(total);
   return(NULL);
  }
  len=strlen(buffer);
  for (len=0;len<size;len++)
   if (buffer[len]==0)
    break;
  total=(char *) realloc(total,len+totalSize+1);
  size*=2;
  buffer=(char *) realloc(buffer,size);
  totalSize+=len;
  strncat(total,buffer,len);
  if (buffer[len-1]=='\n')
   done=true;
 }
 free(buffer);
 return(total);
}

/*************************************************************************
 Description: Reads in a haplotype file.
 Requires: the input file is formatted as described in the documentation
 and has been correctly opened for reading
 Input: f, a pointer to the haplotype input file
 Output: haps, the set of haplotypes described in the file
 Notes: It might be better to implement this with a C++ iostream instead.
**************************************************************************/
void readHaplotypes (FILE *f, vector<Haplotype> &haps)
{
 char *line;
 unsigned total=0;
 unsigned hapSize=16;
 int hapWidth=-1;

 if (f==NULL)
 {
  cerr << "Invalid file handle in readData\n";
  exit(4);
 }
 haps.resize(hapSize);
 while ((line=getOneLine(f)) != NULL)
 {
  int count;
  char *seq=(char *) malloc(strlen(line)*sizeof(char));
  int args;

  args=sscanf(line,"%d %s",&count,seq);
  if (args!=2)
   break;
  if (hapWidth<0)
   hapWidth=strlen(seq);
  else if (((int) strlen(seq))!=hapWidth)
  {
   cerr << "All haplotypes must have the same length\n";
   exit(5);
  }
  haps[total++]=Haplotype(string(seq),count);
  if (total==hapSize)
  {
   hapSize*=2;
   haps.resize(hapSize);
  }
  free(line);
  free(seq);
 }
 haps.resize(total);
}

/*************************************************************************
 Description: Reads in a motif file.
 Requires: the input file is formatted as described in the documentation
 and has been correctly opened for reading
 Input: f, a pointer to the motif input file
 Output: motifProbs, the set of motifs described in the file mapped to
 their associated log frequencies
 Notes: It might be better to implement this with a C++ iostream instead.
**************************************************************************/
void readMotifs (FILE *f, map<Motif,double> &motifProbs)
{
 char *line;

 motifProbs=map<Motif,double>();
 if (f==NULL)
 {
  cerr << "Invalid file handle in readData\n";
  exit(4);
 }
 while ((line=getOneLine(f)) != NULL)
 {
  int start, args;
  char *seq=(char *) malloc(strlen(line)*sizeof(char));
  double freq;
  
  args=sscanf(line,"%d %s %lf",&start,seq,&freq);
  if (args!=3)
   break;
  if (freq<=0.0)
   motifProbs[Motif(start,string(seq))]=-1e12;
  else
   motifProbs[Motif(start,string(seq))]=log(freq);
  free(line);
  free(seq);
 }
}

/*************************************************************************
 Description: Test whether a haplotype string contains any ambiguous
 characters (making it unsuitable for use as a motif).
 Input: s, the string
 Returns: true if the haplotype contains an ambiguous symbol
 (defined here to be '?', 'x', or 'X') and false otherwise
**************************************************************************/
bool ambiguous (string s)
{
 unsigned i;

 for (i=0;i<s.size();i++)
  if (s[i]=='?' || s[i]=='x' || s[i]=='X')
   return(true);

 return(false);
}

/*************************************************************************
 Description: Given a measured binomial trial probability, a number of
 trials, and a number of standard deviations corresponding to a confidence
 interval of a standard normal, find the upper confidence bound of the
 binomial with the given confidence interval.
 Input: oldProb, the measured binomial probability; count, the number of
 trials; devs, the number of standard deviations
 Returns: The log of upper bound of the confidence region on the binomial.
**************************************************************************/
double upperBound (double oldProb, int count, double devs)
{
 double p=exp(oldProb); 

 // Estimate the upper bound using the formula of Agresti and Coull (1998)
 p=(p+devs*devs/(2*count)+devs*sqrt(p*(1-p)/count+devs*devs/(4*count*count)))/
   (1+devs*devs/count);

 // This formula should not be able to produce values above 1, but might
 // as well be sure.
 if (p>=1.0)
   return(0);
 else
  return(log(p));
}

/*************************************************************************
 Description: Given a measured binomial trial probability, a number of
 trials, and a number of standard deviations corresponding to a confidence
 interval of a standard normal, find the upper confidence bound of the
 binomial with the given confidence interval.
 Input: oldProb, the measured binomial probability; count, the number of
 trials; devs, the number of standard deviations
 Returns: The log of upper bound of the confidence region on the binomial.
**************************************************************************/
double lowerBound (double oldProb, int count, double devs)
{
 double p=exp(oldProb); 

 // Estimate the lower bound using the formula of Agresti and Coull (1998)
 p=(p+devs*devs/(2*count)-devs*sqrt(p*(1-p)/count+devs*devs/(4*count*count)))/
   (1+devs*devs/count);

 // This formula should not be able to produce values below 0, but might
 // as well be sure.
 if (p<=0.0)
   return(-1e12);
 else
  return(log(p));
}

/*************************************************************************
 Description: Provide the R part of a mapping from integer colors to RGB
 color values.
 Input: index, an integer color value
 Returns: an integer value from [0,255] corresponding to the R part of an
 RGB value
 Notes: This function should be something that maps distinct low numbers
 to reasonably different RGB values.
**************************************************************************/
inline unsigned char colorR (int index)
{
 return((index*59)%255);
}

/*************************************************************************
 Description: Provide the G part of a mapping from integer colors to RGB
 color values.
 Input: index, an integer color value
 Returns: an integer value from [0,255] corresponding to the G part of an
 RGB value
 Notes: This function should be something that maps distinct low numbers
 to reasonably different RGB values.
**************************************************************************/
inline unsigned char colorG (int index)
{
 return((index*127)%255);
}

/*************************************************************************
 Description: Provide the B part of a mapping from integer colors to RGB
 color values.
 Input: index, an integer color value
 Returns: an integer value from [0,255] corresponding to the B part of an
 RGB value
 Notes: This function should be something that maps distinct low numbers
 to reasonably different RGB values.
**************************************************************************/
inline unsigned char colorB (int index)
{
 return((index*181)%255);
}

/*************************************************************************
 Description: Map a number from [0,255] to a hexadecimal string
 Input: c, a number
 Returns: a string that is the hexadecimal representation of c
**************************************************************************/
inline string hex (unsigned char c)
{
 char a[3];

 sprintf(a,"%x",c);
 if (strlen(a)==1)
  return(string("0")+a);
 else
  return(string(a));
}

/*************************************************************************
 Description: Output the coloring of a set of haplotypes as a raw
 portable anymap (PNM) file.
 Input: haps, a set of haplotypes; colors, the coloring of the
 haplotypes
 Output: prints to standard output a PNM file which displays the coloring
 of the haplotypes
**************************************************************************/
void outputHaplotypesPNM (vector<Haplotype> &haps,
 vector<vector<int> > &colors)
{
 unsigned height=0,width;
 unsigned i,j,k,l;
 unsigned xscale=400,yscale=400;

 for (i=0;i<haps.size();i++)
  height+=(unsigned) haps[i].count;

 if (haps.size()==0)
 {
  cerr << "Cannot output an empty haplotype set\n";
  exit(24);
 }
 else
  width=colors[0].size();

 xscale/=width;
 if (xscale<1)
  xscale=1;
 yscale/=height;
 if (yscale<1)
  yscale=1;

 // Output header
 cout << "P6\n" << width*xscale << " " << height*yscale << "\n255\n";

 // Output data
 for (i=0;i<haps.size();i++)
  if (colors[i].size()!=width)
  {
   cerr << "All haplotypes must have the same length\n";
   exit(11);
  }
  else
   for (k=0;k<yscale*((unsigned) haps[i].count);k++)
    for (j=0;j<width;j++)
     for (l=0;l<xscale;l++)
      cout << colorR(colors[i][j]) << colorG(colors[i][j]) <<
	colorB(colors[i][j]);
}

/*************************************************************************
 Description: Pads an integer number out to a specified length
 with leading spaces and returns the resulting string.
 Input: num, the number; length, the length
 Returns: the respresentation of num as a text string with sufficient
 leading spaces to bring its length to at least length
**************************************************************************/
string pad (int num, unsigned length)
{
 string out;
 char buffer[128];
 unsigned len=0;
 int tmp=num;

 sprintf(buffer,"%d",num);
 if (num<=0)
  len=1;
 while (tmp!=0)
 {
  tmp/=10;
  len++;
 }
 while (len<length)
 {
  out+=" ";
  len++;
 }

 return(out+buffer);
}

/*************************************************************************
 Description: Output the coloring of a set of haplotypes as formatted
 text.
 Input: haps, a set of haplotypes; colors, the coloring of the
 haplotypes
 Output: prints to standard output a text file which displays the
 haplotypes and their coloring
**************************************************************************/
void outputHaplotypesTXT (vector<Haplotype> &haps,
 vector<vector<int> > &colors)
{
 unsigned height,width;
 unsigned i,j;

 height=haps.size();
 if (haps.size()==0)
  width=0;
 else
  width=colors[0].size();

 // Output data
 for (i=0;i<height;i++)
  if (colors[i].size()!=width)
  {
   cerr << "All haplotypes must have the same length\n";
   exit(11);
  }
  else
  {
   // First output the sequence
   cout << "  " << haps[i].sequence[0];
   for (j=1;j<width;j++)
    cout << "   " << haps[i].sequence[j];
   cout << "\n";
   // Now output the colors
   cout << pad(colors[i][0],3);
   for (j=1;j<width;j++)
    cout << pad(colors[i][j],4);
   cout << "\n\n";
  }
}

/*************************************************************************
 Description: Output the coloring of a set of haplotypes as HTML.
 Input: haps, a set of haplotypes; colors, the coloring of the
 haplotypes
 Output: prints to standard output an HTML file which displays the
 haplotypes and their coloring
**************************************************************************/
void outputHaplotypesHTML (vector<Haplotype> &haps,
 vector<vector<int> > &colors)
{
 unsigned height,width;
 unsigned i,j;

 height=haps.size();
 if (haps.size()==0)
  width=0;
 else
  width=colors[0].size();

 // Output header
 cout << "<html>\n<body bgcolor=\"#FFFFFF\">\n";

 // Output data
 for (i=0;i<height;i++)
  if (colors[i].size()!=width)
  {
   cerr << "All haplotypes must have the same length\n";
   exit(11);
  }
  else
  {
   cout << "<code>";
   for (j=0;j<width;j++)
   {
    cout << "<font color=\"#" << hex(colorR(colors[i][j])) <<
     hex(colorG(colors[i][j])) << hex(colorB(colors[i][j])) << "\">" <<
     haps[i].sequence[j] << "</font> ";
   }
   cout << "</code> x " << haps[i].count << " <br>\n";
  }

 // Output trailer
 cout << "</body>\n</html>\n";
}

/*************************************************************************
 Description: Finds a number of standard deviations of a standard normal
 to give a specific confidence interval.
 Input: confidence, the size of the confidence interval
 Returns: the number of standard deviations, or 0 if the confidence
 interval is below 10^-12.
 Note: This method uses a numerical integration strategy that involves
 integrating from -20 standard deviations forward until the necessary
 amount of density has been accumulated.  It will not work for confidence
 intervals requiring more than 20 standard deviations.
**************************************************************************/
double sigmaFromInterval (double confidence)
{
 double p=(1-confidence)/2;
 double f=0.0,x=-20.0;
 const double tol=0.01, c=sqrt(2*3.14159265358979323846264);

 if (confidence<1e-12)
  return(0.0);

 // Find normal deviate by numerical integration
 do
 {
  double dx,df,dxd;

  // Pick a maximum step size
  dx=1.0;
  while (dx>-p*tol/(10.0*dx*(x+dx))*exp((x+dx)*(x+dx)/2))
   dx/=2.0;

  // Pick a desired step size
  dxd=(p-f)*c*exp(x*x/2);

  if (dxd<=dx)
   dx=dxd;

  df=dx*exp(-x*x/2)/c;

  x+=dx;
  f+=df;

 } while (fabs(p-f)>p*tol);

 // Return the number of standard deviations
 return(-x);
}
